import logging
import os
from typing import List
import proto_schema_parser.parser as pb_parser
from proto_schema_parser.ast import Package, Service, Message, Method, Field

from appsec_discovery.parsers import Parser
from appsec_discovery.models import CodeObject, CodeObjectField

logger = logging.getLogger(__name__)

class ProtobufParser(Parser):

    def find_proto_files(self, root_dir):
        proto_files = []
        for root, _, files in os.walk(root_dir):
            for file in files:
                if file.endswith('.proto'):
                    proto_files.append(os.path.join(root, file))
        return sorted(proto_files)


    def run_scan(self) -> List[CodeObject]:

        objects_list: List[CodeObject] = []

        proto_files = self.find_proto_files(self.source_folder)
        proto_data = {}

        for proto_file in proto_files:

            local_proto_file = proto_file.replace(self.source_folder, "")

            try:
                with open(proto_file) as file:
                    file_str = file.read()
                    proto_data[local_proto_file] = pb_parser.Parser().parse(file_str)
            except Exception as ex:
                logger.error(f"Failed to parse {self.parser} for file {local_proto_file}: {ex}")

        objects_list = self.parse_report(proto_data)

        return objects_list

    def resolve_fields(self, type_name, messages, file, depth):
        
        resolved_fields = {}

        local_messages = {}

        if depth > 5 :
            return {}

        for el in messages[type_name].elements:       
            if isinstance(el, Message) and el.name not in local_messages:
                local_messages[el.name] = el

        for el in messages[type_name].elements:      
            if isinstance(el, Field) and el.name not in resolved_fields:

                if el.type in messages:
                    resolved_local_fields = self.resolve_fields(el.type, messages, file, depth + 1)

                    for field_name, field in resolved_local_fields.items():
                        resolved_fields[f"{type_name}.{el.name}.{field_name}"] = field

                elif el.type in local_messages:
                    resolved_local_fields = self.resolve_fields(el.type, local_messages, file, depth + 1)

                    for field_name, field in resolved_local_fields.items():
                        resolved_fields[f"{type_name}.{el.name}.{field_name}"] = field

                else:
                    resolved_fields[f"{type_name}.{el.name}"] = CodeObjectField(
                        field_name=f"{type_name}.{el.name}",
                        field_type=el.type,
                        file=file,
                        line=el.number
                    )

        if not resolved_fields :
            resolved_fields[type_name] = CodeObjectField(
                        field_name=type_name,
                        field_type='Empty',
                        file=file,
                        line=1
                    )

        return resolved_fields

    def parse_report(self, proto_data) -> List[CodeObject]:

        parsed_objects: List[CodeObject] = []

        packages = {}

        for file, file_proto in proto_data.items():

            cur_package = ''

            for el in file_proto.file_elements:

                if isinstance(el, Package) :

                    cur_package = el.name

                    if el.name not in packages:
                        packages[el.name] = {
                            'services': {},
                            'messages': {},
                            'file': file
                        }

            for el in file_proto.file_elements:

                if isinstance(el, Service) and el.name not in packages[cur_package]['services']:
                    packages[cur_package]['services'][el.name] = {}

                    for service_method in el.elements:
                        if isinstance(service_method, Method) and service_method.name not in packages[cur_package]['services'][el.name]:
                            packages[cur_package]['services'][el.name][service_method.name] = {
                                'input': service_method.input_type.type,
                                'output': service_method.output_type.type
                            }
                
                if isinstance(el, Message) and el.name not in packages[cur_package]['messages']:
                    packages[cur_package]['messages'][el.name] = el
        

        for package_name, package in packages.items():
            for service_name, service in package['services'].items():
                for method_name, method in service.items():

                    unique_hash = self.calc_uniq_hash([package['file'], package_name, service_name, method_name])

                    code_object = CodeObject(
                        hash=unique_hash,
                        object_name=f"Rpc /{package_name}.{service_name}/{method_name}",
                        object_type='rpc',
                        parser=self.parser,
                        file=package['file'],
                        line=1,
                        properties={},
                        fields={}
                    )

                    if method['input'] in package['messages']:
                        input_fields = self.resolve_fields(method['input'], package['messages'], package['file'], 0)

                        for field_name, field in input_fields.items():    
                            code_object.fields[f"input.{field_name}"] = field
                    else:
                        code_object.fields['input'] = CodeObjectField(
                            field_name='input',
                            field_type=method['input'],
                            file=package['file'],
                            line=1
                        )

                    if method['output'] in package['messages']:
                        output_fields = self.resolve_fields(method['output'], package['messages'], package['file'], 0)

                        for field_name, field in output_fields.items():    
                            code_object.fields[f"output.{field_name}"] = field
                    else:
                        code_object.fields['output'] = CodeObjectField(
                            field_name='output',
                            field_type=method['output'],
                            file=package['file'],
                            line=1
                        )

                    parsed_objects.append(code_object)

        logger.info(f"For scan {self.parser} data parse {len(parsed_objects)} objects")

        return parsed_objects


from llama_cpp import Llama, LlamaRAMCache
from openai import OpenAI

import re

from typing import List, Dict
import logging

from appsec_discovery.models import CodeObject, ExcludeScoring, AiLocal, AiApi

severities_int = {'critical': 5, 'high': 4, 'medium': 3, 'low': 2, 'info': 1}
skip_ai = ['created_at', 'updated_at', 'deleted_at']

logger = logging.getLogger(__name__)

class AiService:

    def __init__(self, exclude_scoring: List[ExcludeScoring], ai_local: AiLocal = None, ai_api: AiApi = None):

        self.ai_local = ai_local
        self.ai_api = ai_api

        self.exclude_scoring = exclude_scoring       

    def ai_score_objects(self, code_objects: List[CodeObject]) -> List[CodeObject]:

        scored_objects: List[CodeObject] = []

        try:

            for object in code_objects:

                object_to_score: Dict[str: List[str]] = {'field_names': []}
                
                scored_list = {}

                for field in object.fields.values() :

                    if field.field_name.split('.')[0].lower() in ['input','output'] and len(field.field_name.split('.')) > 1 :
                        field_name = ".".join(field.field_name.split('.')[1:])
                    else:
                        field_name = field.field_name

                    question = f'''
                        For object: {object.object_name}
                        
                        Field name: {field_name}

                        Can contain private data? Answer only 'yes' or 'no',
                    '''

                    question2 = f'''
                        For object: {object.object_name}
                        
                        Field name: {field_name}

                        Choose category for field private data from lost: pii, finance, auth, other 

                        Answer only with category name word.
                    '''
                    
                    # Local
                    if self.ai_local:

                        llm = Llama.from_pretrained(
                            repo_id=self.ai_local.model_id,
                            filename=self.ai_local.gguf_file,
                            verbose=False,
                            cache_dir=self.ai_local.model_folder,
                            max_tokens=5,
                            seed=112358,
                        )

                        response = llm.create_chat_completion(
                            messages = [
                                {"role": "system", "content": self.ai_local.system_prompt},
                                {"role": "user", "content": question },
                            ]
                        )

                        llm.reset()
                        llm.set_cache(None)

                        answer = response['choices'][0]["message"]["content"]

                        if 'yes' in answer.lower():

                            response2 = llm.create_chat_completion(
                                messages = [
                                    {"role": "system", "content": self.ai_local.system_prompt},
                                    {"role": "user", "content": question2 },
                                ]
                            )

                            answer2 = response2['choices'][0]["message"]["content"]

                            for cat in ['pii', 'auth', 'finance', 'other']:
                                if cat in answer2.lower():
                                    scored_list[field.field_name] = f"llm-{cat}"
                                    logger.info(f"For {object.object_name} and {field.field_name} llm answer is {answer}, cat {answer2}")

                    # API
                    if self.ai_api:

                        client = OpenAI(api_key=self.ai_api.api_key, base_url=self.ai_api.base_url)

                        response = client.chat.completions.create(
                            model=self.ai_api.model,
                            messages=[
                                {"role": "system", "content": self.ai_api.system_prompt},
                                {"role": "user", "content": question},
                            ],
                            stream=False
                        )

                        answer = response.choices[0].message.content

                        if 'yes' in answer.lower():

                            response2 = client.chat.completions.create(
                                model=self.ai_api.model,
                                messages=[
                                    {"role": "system", "content": self.ai_api.system_prompt},
                                    {"role": "user", "content": question2},
                                ],
                                stream=False
                            )

                            answer2 = response2.choices[0].message.content

                            for cat in ['pii', 'auth', 'finance', 'other']:
                                if cat in answer2.lower():
                                    scored_list[field.field_name] = f"llm-{cat}"
                                    logger.info(f"For {object.object_name} and {field.field_name} llm answer is {answer}, cat {answer2}")
                
                scored_fields = {}

                severity = "medium"

                for field_name, field in object.fields.items():

                    if field.field_name in scored_list.keys():

                        excluded = False

                        tag = scored_list[field.field_name]

                        for exclude in self.exclude_scoring:

                            if ( exclude.file or exclude.parser or exclude.object_name or exclude.object_type or exclude.prop_name 
                                    or exclude.field_name or exclude.field_type or exclude.tag or exclude.keyword ) \
                                and (exclude.parser is None or exclude.parser.lower() == object.parser.lower()) \
                                and (exclude.file is None or re.match(exclude.file, object.file) or exclude.file.lower() in object.file.lower()) \
                                and (exclude.object_name is None or re.match(exclude.object_name, object.object_name) or exclude.object_name.lower() in object.object_name.lower()) \
                                and (exclude.object_type is None or re.match(exclude.object_type, object.object_type) or exclude.object_type.lower() in object.object_type.lower()) \
                                and (exclude.prop_name is None ) \
                                and (exclude.field_name is None or re.match(exclude.field_name, field.field_name) or exclude.field_name.lower() in field.field_name.lower()) \
                                and (exclude.field_type is None or re.match(exclude.field_type, field.field_type) or exclude.field_type.lower() in field.field_type.lower()) \
                                and (exclude.tag is None or exclude.tag == tag ) \
                                and (exclude.keyword is None ) :
                                
                                excluded = True

                        if not excluded :

                            if not field.severity:
                                field.severity = severity
                                field.tags = [tag]
                            else:
                                if tag not in field.tags:
                                    field.tags.append(tag)

                                if severities_int[severity] > severities_int[field.severity]:
                                    field.severity = severity

                            if not object.severity:
                                object.severity = severity
                                object.tags = [tag]
                            else:
                                if tag not in object.tags:
                                    object.tags.append(tag)

                                if severities_int[severity] > severities_int[object.severity]:
                                    object.severity = severity

                    scored_fields[field_name] = field
                
                object.fields = scored_fields

                scored_objects.append(object)

        except Exception as ex:
            logger.error(f"Error while ai scoring: {ex}")

        # Check that all processed
        if len(scored_objects) == len(code_objects):
            return scored_objects
        
        else:
            return code_objects



import logging
import os
from typing import List
from pathlib import Path
from openapi_parser import parse

from appsec_discovery.parsers import Parser
from appsec_discovery.models import CodeObject, CodeObjectProp, CodeObjectField

logger = logging.getLogger(__name__)

class SwaggerParser(Parser):

    def find_swagger_files(self, root_dir):

        swagger_files = []
        for root, _, files in os.walk(root_dir):
            for file in files:
                if file.endswith('.yaml') or file.endswith('.yml') or file.endswith('.json') :
                    try:
                        with open(os.path.join(root, file)) as file_obj:
                            file_str = file_obj.read()
                            if 'openapi' in file_str and 'paths' in file_str:
                                swagger_files.append(os.path.join(root, file))
                    except:
                        pass

        return sorted(swagger_files)


    def run_scan(self) -> List[CodeObject]:

        objects_list: List[CodeObject] = []

        swagger_files = self.find_swagger_files(self.source_folder)
        swagger_data = {}

        for swagger_file in swagger_files:

            local_swagger_file = swagger_file.replace(self.source_folder, "")

            try:
                swagger_data[local_swagger_file] = parse(swagger_file)
            except:
                pass
            
        objects_list = self.parse_report(swagger_data)

        return objects_list

    def resolve_fields(self, type, object, file, depth):
        
        resolved_fields = {}

        if depth > 10:
            return resolved_fields

        # Object
        if hasattr(object, 'properties'):

            for prop in object.properties:

                if not hasattr(prop.schema, 'items') and not hasattr(prop.schema, 'properties'):

                    resolved_fields[f"{type}.{prop.name}"] = CodeObjectField(
                        field_name=f"{type}.{prop.name}",
                        field_type=prop.schema.type.value,
                        file=file,
                        line=1
                    )
                
                else:

                    res_fields = self.resolve_fields(prop.name, prop.schema, file, depth + 1)

                    for field_name, field_data in res_fields.items():
                        resolved_fields[f"{type}.{field_name}"] = field_data
                        resolved_fields[f"{type}.{field_name}"].field_name = f"{type}.{field_name}"

        # List 
        elif hasattr(object, 'items'):

            res_fields = self.resolve_fields('', object.items, file, depth + 1)

            for field_name, field_data in res_fields.items():
                field_name = field_name.strip('.')
                resolved_fields[f"{type}.{field_name}"] = field_data
                resolved_fields[f"{type}.{field_name}"].field_name = f"{type}.{field_name}"
            
        else:
            resolved_fields[type] = CodeObjectField(
                field_name=type,
                field_type=object.type.value,
                file=file,
                line=1
            )

        return resolved_fields

    def parse_report(self, swagger_data) -> List[CodeObject]:

        parsed_objects: List[CodeObject] = []

        for file, file_spec in swagger_data.items():

            for path in file_spec.paths :

                for method in path.operations:

                    unique_hash = self.calc_uniq_hash([method.method.name, path.url, file])

                    code_object = CodeObject(
                        hash=unique_hash,
                        object_name=f"Route {path.url} ({method.method.name})",
                        object_type='route',
                        parser=self.parser,
                        file=file,
                        line=1,
                        properties={},
                        fields={}
                    )

                    code_object.properties['path'] = CodeObjectProp(
                        prop_name='path',
                        prop_value=path.url
                    )

                    code_object.properties['method'] = CodeObjectProp(
                        prop_name='method',
                        prop_value=method.method.name
                    )

                    for param in method.parameters:

                        code_object.fields[f"{param.location.value}.param.{param.name}"] = CodeObjectField(
                            field_name=f"{param.location.value}.param.{param.name}",
                            field_type=param.schema.type.value,
                            file=file,
                            line=1
                        )

                    if method.request_body :
                        for content in method.request_body.content :

                            res = self.resolve_fields('input', content.schema, file, 0)

                            for field_name, field in res.items():
                                code_object.fields[field_name] = field

                    if method.responses :
                        for response in method.responses :
                            if response.content:
                                for content in response.content :

                                    res = self.resolve_fields('output', content.schema, file, 0)

                                    for field_name, field in res.items():
                                        code_object.fields[field_name] = field

                    parsed_objects.append(code_object)
            
        logger.info(f"For scan {self.parser} data parse {len(parsed_objects)} objects")

        return parsed_objects